/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifndef AIRPORT_H_
#define AIRPORT_H_

#include <stdint.h>
#include <string>
#include <vector>
#include "Waypoint.h"
#include "Runway.h"
#include "NavAid.h"
#include "Frequency.h"
#include "ApproachPlate.h"
#include "METAR.h"

class Airport : public Waypoint
{
public:
	Airport(std::string IATA="",std::string ICAO="");
	~Airport();

	bool ParseNFD(std::string Line);
	bool ParseXPlane850(std::string *Line);

	std::string GetIATA();
	std::string GetICAO();
	double GetMagVar();
	int GetElevation();
	std::vector<Frequency*> *GetFreqs();
	std::vector<Runway*> *GetRunways();
	std::vector<NavAid*> *GetNavAids();
	std::vector<ApproachPlate*> *GetApproachPlates();
	std::vector<METAR*> *GetMETARs();
	void AddRunway(Runway *R);
	void AddNavAid(NavAid *N);
	void AddFrequency(Frequency *F);
	void AddApproachPlate(ApproachPlate *A);
	void AddMETAR(METAR *M);
	void ClearApproachPlates();
	void ClearMETARs();
	bool MakeBinary(struct Airport_Binary_struct *BinaryData);
	void SetFromBinary(struct Airport_Binary_struct *BinaryData);
	bool Matchi(std::string str);
	static bool IndirectLessThan (Airport* left, Airport *right);
	void ComputeLocByRunways();

private:
	std::string IATA;
	std::string ICAO;
	std::vector<Runway*> Runways;
	std::vector<NavAid*> NavAids;
	std::vector<Frequency*> Freqs;
	std::vector<ApproachPlate*> ApproachPlates;
	double MagVar;	/* Degrees */
	int Elevation;	/* Feet */
	std::vector<METAR*> METARs;
};

/* Increment the version when the structure changes */
#define AIRPORT_BINARY_STRUCT_VER 1
#ifdef WIN32
#pragma pack(1)
#endif
struct Airport_Binary_struct {
	/* Runway, NavAid, Frequency, and Approach Plate info is not included since they
	 * may occur variable numbers of times
	 */
	char ID[5];
	char Name[50];
	char IATA[3];
	char ICAO[4];
	int32_t Lonx1000000,Latx1000000;
	int16_t MagVarx100;
	int16_t Elevation;
	unsigned char nRunways,nNavAids,nFreqs,nApproachPlates;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif


bool operator < (const Airport& left, const Airport& right);
bool operator == (const Airport& left, const Airport& right);

#endif /*AIRPORT_H_*/
