/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <vector>
#include <glibmm/i18n.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include <gtkmm/treemodel.h>
#include "ChartChooserDialog.h"
#include "ChartChooserDrawingArea.h"
#include "Exception.h"
#include "GeoChart.h"

ChartChooserDialog::ChartChooserDialog(std::string *GladeFilename, Config *Cfg, DataStore *Data, Route *RouteData)
{
	try
	{
		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ChartChooserDialog");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog(_("Error reading data file: ")+*GladeFilename+"\n"+_("Program cannot continue."),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
		return;
	}

	refXml->get_widget("ChartChooserDialog",GtkDialog);
	refXml->get_widget_derived("HawaiiDrawingArea",HawaiiDrawingArea);
	refXml->get_widget_derived("AlaskaDrawingArea",AlaskaDrawingArea);
	refXml->get_widget_derived("ContiguousDrawingArea",ContiguousDrawingArea);
	refXml->get_widget("SectionalsRadioButton",SectionalsRadioButton);
	refXml->get_widget("TACsRadioButton",TACsRadioButton);
	refXml->get_widget("FlywaysRadioButton",FlywaysRadioButton);
	refXml->get_widget("FlywayComboBox",FlywayComboBox);
	refXml->get_widget("SelectedLabel",SelectedLabel);

	GtkDialog->add_events(Gdk::BUTTON_PRESS_MASK);
	GtkDialog->add_events(Gdk::BUTTON_RELEASE_MASK);
	GtkDialog->signal_button_release_event().connect(sigc::mem_fun(this,&ChartChooserDialog::on_button_release_event));
	SectionalsRadioButton->signal_clicked().connect(sigc::mem_fun(this,&ChartChooserDialog::on_Sectionals_clicked));
	TACsRadioButton->signal_clicked().connect(sigc::mem_fun(this,&ChartChooserDialog::on_TACs_clicked));
	FlywaysRadioButton->signal_clicked().connect(sigc::mem_fun(this,&ChartChooserDialog::on_Flyways_clicked));
	FlywayComboBox->signal_changed().connect(sigc::mem_fun(this,&ChartChooserDialog::on_FlywayComboBox_changed));

	this->Data=Data;
	this->Cfg=Cfg;

	/* Set the current selection and hover status to invalid */
	SelectedCharti=Data->GetGeoCharts()->end();

	HawaiiDrawingArea->Init(std::string(PACKAGEDATADIR)+"/Hawaii.png",Data,ChartChooserDrawingArea::AREA_HAWAII,&SelectedCharti,RouteData);
	AlaskaDrawingArea->Init(std::string(PACKAGEDATADIR)+"/Alaska.png",Data,ChartChooserDrawingArea::AREA_ALASKA,&SelectedCharti,RouteData);
	ContiguousDrawingArea->Init(std::string(PACKAGEDATADIR)+"/ContiguousUSA.png",Data,ChartChooserDrawingArea::AREA_CONTIGUOUS,&SelectedCharti,RouteData);

	/* Start with Sectionals selected */
	on_Sectionals_clicked();
	
	/* Setup the Flyway ComboBox */
	FlywayComborefTreeModel = Gtk::ListStore::create(FlywayComboColumns);
	FlywayComboBox->set_model(FlywayComborefTreeModel);
	
	for (std::vector<GeoChart*>::iterator GeoCharti=Data->GetGeoCharts()->begin();GeoCharti!=Data->GetGeoCharts()->end();++GeoCharti) {
		if ((*GeoCharti)->GetType()==GeoChart::TYPE_FLYWAY) {
			Gtk::TreeModel::Row row = *(FlywayComborefTreeModel->append());
			row[FlywayComboColumns.Name] = (*GeoCharti)->GetName();
			row[FlywayComboColumns.Charti] = GeoCharti;
		}
	}
	FlywayComboBox->pack_start(FlywayComboColumns.Name);

	/* Since the size of the DrawingAreas aren't known until runtime, recenter the dialog here */
	GtkDialog->set_position(Gtk::WIN_POS_CENTER_ALWAYS);
}

ChartChooserDialog::~ChartChooserDialog()
{
	delete GtkDialog;
}

int ChartChooserDialog::Run() {
	return GtkDialog->run();
}

void ChartChooserDialog::on_Sectionals_clicked() {
	ContiguousDrawingArea->SetChartType(GeoChart::TYPE_SECTIONAL);
	AlaskaDrawingArea->SetChartType(GeoChart::TYPE_SECTIONAL);
	HawaiiDrawingArea->SetChartType(GeoChart::TYPE_SECTIONAL);
	ContiguousDrawingArea->Redraw();
	AlaskaDrawingArea->Redraw();
	HawaiiDrawingArea->Redraw();
	FlywayComboBox->set_sensitive(false);
	
}

void ChartChooserDialog::on_TACs_clicked() {
	ContiguousDrawingArea->SetChartType(GeoChart::TYPE_TAC);
	AlaskaDrawingArea->SetChartType(GeoChart::TYPE_TAC);
	HawaiiDrawingArea->SetChartType(GeoChart::TYPE_TAC);
	ContiguousDrawingArea->Redraw();
	AlaskaDrawingArea->Redraw();
	HawaiiDrawingArea->Redraw();
	FlywayComboBox->set_sensitive(false);
}

void ChartChooserDialog::on_Flyways_clicked() {
	ContiguousDrawingArea->SetChartType(GeoChart::TYPE_FLYWAY);
	AlaskaDrawingArea->SetChartType(GeoChart::TYPE_FLYWAY);
	HawaiiDrawingArea->SetChartType(GeoChart::TYPE_FLYWAY);
	SelectedCharti=Data->GetGeoCharts()->end();
	ContiguousDrawingArea->Redraw();
	AlaskaDrawingArea->Redraw();
	HawaiiDrawingArea->Redraw();
	FlywayComboBox->set_sensitive(true);
	on_FlywayComboBox_changed();	/* Update the display as necessary */
}

bool ChartChooserDialog::on_button_release_event(GdkEventButton * event) {
	ContiguousDrawingArea->Redraw();
	AlaskaDrawingArea->Redraw();
	HawaiiDrawingArea->Redraw();
	SetSelectedLabel();
	return true;
}

void ChartChooserDialog::on_FlywayComboBox_changed() {
	Gtk::TreeModel::iterator Activei;
	Activei=FlywayComboBox->get_active();
	if (Activei) {		/* If anything is selected */
		Gtk::TreeModel::Row row = *Activei;
		SelectedCharti=row[FlywayComboColumns.Charti];
	} else {
		/* Nothing selected */
		SelectedCharti=Data->GetGeoCharts()->end();
	}

	SetSelectedLabel();
}

void ChartChooserDialog::SetSelectedLabel() {
	if (SelectedCharti==Data->GetGeoCharts()->end()) {
		SelectedLabel->set_text("");
	} else {
		SelectedLabel->set_text((*SelectedCharti)->GetName());
	}
}

std::vector<GeoChart*>::iterator ChartChooserDialog::GetSelectedChart() {
	return SelectedCharti;
}
