/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <fstream>
#include <iostream>
#include <sstream>
#include <ios>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <cstring>
#include <libglademm.h>
#include <glibmm/i18n.h>
#include <gtkmm/window.h>
#include <gtkmm/progressbar.h>
#include <gtkmm/button.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include "DataStore.h"
#include "Exception.h"
#include "CRC32.h"
#include "Util.h"

Volume::Volume(std::string *GladeFilename, Config *Cfg)
{
	this->GladeFilename=GladeFilename;
	this->Cfg=Cfg;

	this->Name="";
	this->EffectiveFrom=0;
	this->EffectiveTo=0;
	nAirports=0;
	nNavAids=0;
	nEnrouteWaypoints=0;
	nGeoCharts=0;
	nApproachPlates=0;

	Loaded=false;
}

Volume::Volume(std::string *GladeFilename, Config *Cfg, std::string Name, time_t EffectiveFrom, time_t EffectiveTo)
{
	this->GladeFilename=GladeFilename;
	this->Cfg=Cfg;

	this->Name=Name;
	this->EffectiveFrom=EffectiveFrom;
	this->EffectiveTo=EffectiveTo;
	nAirports=0;
	nNavAids=0;
	nEnrouteWaypoints=0;
	nGeoCharts=0;
	nApproachPlates=0;

	Loaded=false;

	/* Create the dest dir
	 * Ignore errors since it's not really an error for the directory to already exist
	 * and anything else will get caught with the file copy
	 */
	std::string Dir;
	Dir=Cfg->GetDataStoreDir()+"/"+Name;
	Util::ShellExpand(&Dir);
	Util::MkDir(Dir,0775);
	this->Filename=Dir+"/volume.vifir";
}

Volume::~Volume()
{
	Clear();
}

void Volume::Clear()
{
	ClearFlightDB();
	ClearGeoCharts();

	Loaded=false;
}

void Volume::ClearFlightDB()
{
	ClearApproachPlates();
	for(std::vector<Airport*>::iterator Airporti = Airports.begin(); Airporti != Airports.end(); ++Airporti)
	{
		delete *Airporti;
	}
	Airports.clear();

	for(std::vector<NavAid*>::iterator NavAidi = NavAids.begin(); NavAidi != NavAids.end(); ++NavAidi)
	{
		delete *NavAidi;
	}
	NavAids.clear();

	for(std::vector<EnrouteWaypoint*>::iterator EnrouteWaypointi = EnrouteWaypoints.begin(); EnrouteWaypointi != EnrouteWaypoints.end(); ++EnrouteWaypointi)
	{
		delete *EnrouteWaypointi;
	}
	EnrouteWaypoints.clear();
}

void Volume::ClearGeoCharts() {
	for(std::vector<GeoChart*>::iterator GeoCharti = GeoCharts.begin(); GeoCharti != GeoCharts.end(); ++GeoCharti)
	{
		delete *GeoCharti;
	}
	GeoCharts.clear();
}

void Volume::ClearApproachPlates() {
	nApproachPlates=GetnApproachPlates();	/* Update our cache in case new approach plates have been added */

	for(std::vector<Airport*>::iterator Airporti = Airports.begin(); Airporti != Airports.end(); ++Airporti)
	{
		(*Airporti)->ClearApproachPlates();
	}
}

std::vector<Airport*> *Volume::GetAirports()
{
	return &Airports;
}

unsigned int Volume::GetnAirports() {
	if (Loaded)
		return Airports.size();
	else
		return nAirports;
}

std::vector<NavAid*> *Volume::GetNavAids()
{
	return &NavAids;
}

unsigned int Volume::GetnNavAids() {
	if (Loaded)
		return NavAids.size();
	else
		return nNavAids;
}

std::vector<EnrouteWaypoint*> *Volume::GetEnrouteWaypoints()
{
	return &EnrouteWaypoints;
}

unsigned int Volume::GetnEnrouteWaypoints() {
	if (Loaded)
		return EnrouteWaypoints.size();
	else
		return nEnrouteWaypoints;
}

std::vector<GeoChart*> *Volume::GetGeoCharts()
{
	return &GeoCharts;
}

unsigned int Volume::GetnGeoCharts() {
	if (Loaded)
		return GeoCharts.size();
	else
		return nGeoCharts;
}

unsigned int Volume::GetnApproachPlates() {
	unsigned int AppCount=0;

	if (Loaded) {
		for (std::vector<Airport*>::iterator Ai=Airports.begin();Ai!=Airports.end();++Ai) {
			AppCount+=(*Ai)->GetApproachPlates()->size();
		}
		return AppCount;
	} else {
		return nApproachPlates;
	}
}

Airport *Volume::GetAirportByICAO(std::string ICAO)
{
	for(std::vector<Airport*>::iterator Airporti = Airports.begin(); Airporti != Airports.end(); ++Airporti) {
		if ((*Airporti)->GetICAO()==ICAO)
		{
			return *Airporti;
		}
	}
	return NULL;
}

Airport *Volume::GetAirportByIATA(std::string IATA)
{
	for(std::vector<Airport*>::iterator Airporti = Airports.begin(); Airporti != Airports.end(); ++Airporti) {
		if ((*Airporti)->GetIATA()==IATA)
		{
			return *Airporti;
		}
	}
	return NULL;
}

/* Str=string to search for, Waypoints=vector to fill with matching waypoints */
void Volume::WaypointSearch(std::string Str, std::vector<Waypoint*> *Waypoints, unsigned int Max)
{
	/* Search airports */
	for(std::vector<Airport*>::iterator Airporti = Airports.begin(); Airporti != Airports.end(); Airporti++) {
		if ((*Airporti)->Matchi(Str))
		{
			Waypoints->push_back((Waypoint*)(*Airporti));
			if (Waypoints->size()>Max) {
				Waypoints->clear();
				return;
			}
		}
	}

	/* Search NavAids */
	for(std::vector<NavAid*>::iterator NavAidi = NavAids.begin(); NavAidi != NavAids.end(); NavAidi++) {
		if ((*NavAidi)->Matchi(Str))
		{
			Waypoints->push_back((Waypoint*)(*NavAidi));
			if (Waypoints->size()>Max) {
				Waypoints->clear();
				return;
			}
		}
	}

	/* Search Enroute Waypoints */
	for(std::vector<EnrouteWaypoint*>::iterator EnrouteWaypointi = EnrouteWaypoints.begin(); EnrouteWaypointi != EnrouteWaypoints.end(); EnrouteWaypointi++) {
		if ((*EnrouteWaypointi)->Matchi(Str))
		{
			Waypoints->push_back((Waypoint*)(*EnrouteWaypointi));
			if (Waypoints->size()>Max) {
				Waypoints->clear();
				return;
			}
		}
	}
}

time_t Volume::GetEffectiveFrom() {
	return EffectiveFrom;
}
time_t Volume::GetEffectiveTo() {
	return EffectiveTo;
}

std::string Volume::GetName() {
	return Name;
}

bool Volume::Load(std::string Filename) {
	std::ifstream ifs;
	struct stat statdata;
	uint32_t RunningCRC,HeaderCRC;
	uint32_t i;
	Glib::RefPtr<Gnome::Glade::Xml> refXml;
	Gtk::Window *ProgressWindow;
	Gtk::ProgressBar *ProgressBar,*SubProgressBar;
	Gtk::Button *CancelButton;
	double ProgressDone=0.0;
	struct Volume_Header_struct Header;

	if (Loaded)
		return true;	/* Already loaded, nothing to do */

	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog(_("Error reading data file: ")+*GladeFilename+"\n"+_("Program cannot continue."),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
		return false;
	}

	refXml->get_widget("ProgressWindow",ProgressWindow);
	refXml->get_widget("SubProgressBar",SubProgressBar);
	refXml->get_widget("ProgressBar",ProgressBar);
	refXml->get_widget("CancelButton",CancelButton);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&Volume::on_CancelButton_clicked));
	CancelFlag=false;

	ProgressWindow->set_deletable(false);
	ProgressBar->set_text("Total");
	ProgressWindow->show_all();

	try {
		if (!Scan(Filename,&Header)) {
			delete ProgressWindow;
			return false;
		}
	}
	catch (Exception Ex) {
		delete ProgressWindow;
		throw Ex;
	}

	/* Set the header filesize and CRC fields to zero and start computing the CRC */
	Header.FileSize=0;
	HeaderCRC=Header.CRC32;
	Header.CRC32=0;
	RunningCRC=CRC.Compute(&Header,sizeof(Header));

	if (stat(this->Filename.c_str(),&statdata)!=0) {
		if (errno==ENOENT) {
			/* A missing file isn't really an error - we'll just
			 * proceed empty
			 */
			delete ProgressWindow;
			return true;
		}

		delete ProgressWindow;
		throw Exception("Error processing "+this->Filename,Exception::LEVEL_ERROR);
	}

	ifs.open(this->Filename.c_str(),std::ifstream::binary);
	ifs.seekg(sizeof(struct Volume_Header_struct),std::ios::beg);	/* Go to just beyond the header */
	if (!ifs.good()) {
		delete ProgressWindow;
		throw Exception("Error opening "+this->Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	/* Discard what's currently loaded */
	Clear();

	/* Load the data */
	try {
		SubProgressBar->set_text("Airports");
		for (i=0;i<Header.nAirports && !CancelFlag;i++) {
			if (i%100==0) {
				SubProgressBar->set_fraction((double)i/Header.nAirports);
				ProgressBar->set_fraction(AirportsProgressFrac*i/Header.nAirports);
				Util::ScheduleMainLoop();
			}
			LoadAirport(&Airports,&ifs,&RunningCRC);
		}
		SubProgressBar->set_text("NavAids");
		ProgressDone=AirportsProgressFrac;
		for (i=0;i<Header.nNavAids && !CancelFlag;i++) {
			if (i%100==0) {
				SubProgressBar->set_fraction((double)i/Header.nNavAids);
				ProgressBar->set_fraction(ProgressDone+NavAidsProgressFrac*i/Header.nNavAids);
				Util::ScheduleMainLoop();
			}
			LoadNavAid(&NavAids,&ifs,&RunningCRC);
		}
		ProgressDone+=NavAidsProgressFrac;
		SubProgressBar->set_text("Waypoints");
		for (i=0;i<Header.nEnrouteWaypoints && !CancelFlag;i++) {
			if (i%100==0) {
				SubProgressBar->set_fraction((double)i/Header.nEnrouteWaypoints);
				ProgressBar->set_fraction(ProgressDone+WaypointsProgressFrac*i/Header.nEnrouteWaypoints);
				Util::ScheduleMainLoop();
			}
			LoadEnrouteWaypoint(&EnrouteWaypoints,&ifs,&RunningCRC);
		}
		SubProgressBar->set_text("Charts");
		ProgressDone+=WaypointsProgressFrac;
		for (i=0;i<Header.nGeoCharts && !CancelFlag;i++) {
			if (i%100==0) {
				SubProgressBar->set_fraction((double)i/Header.nGeoCharts);
				ProgressBar->set_fraction(ProgressDone+ChartsProgressFrac*i/Header.nGeoCharts);
				Util::ScheduleMainLoop();
			}
			LoadGeoChart(&GeoCharts,&ifs,&RunningCRC);
		}

		/* Check the CRC */
		if (HeaderCRC!=CRC.Finalize(RunningCRC) && !CancelFlag) {
			Clear();
			throw Exception("File corrupt - invalid CRC in "+this->Filename,Exception::LEVEL_ERROR);
		}
	} catch (Exception Ex) {
		ifs.close();
		Clear();
		delete ProgressWindow;
		throw Ex;
	}

	ifs.close();
	delete ProgressWindow;

	if (CancelFlag) {
		Clear();
		return false;
	}

	Loaded=true;

	return true;
}

void Volume::Save(std::string Filename) {
	std::ofstream ofs;
	struct Volume_Header_struct Header;

	if (Filename.compare("")!=0) {
		Util::ShellExpand(&Filename);
		this->Filename=Filename;
	} else {
		Filename=this->Filename;
	}

	ofs.open(Filename.c_str(),std::ofstream::binary);
	if (!ofs.good()) {
		throw Exception("Error opening "+Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	memset(&Header,0,sizeof(Header));
	strncpy(Header.Magic,VOLUME_FILEMAGIC,sizeof(VOLUME_FILEMAGIC));
	Header.DataVersion=VOLUME_HEADER_STRUCT_VER;
	strncpy(Header.VolumeName,Name.c_str(),sizeof(Header.VolumeName));
	Header.EffectiveFrom=EffectiveFrom;
	Header.EffectiveTo=EffectiveTo;
	Header.nAirports=Airports.size();
	Header.nNavAids=NavAids.size();
	Header.nEnrouteWaypoints=EnrouteWaypoints.size();
	Header.nGeoCharts=GeoCharts.size();
	Header.nApproachPlates=GetnApproachPlates();
	/* Leave FileSize set to 0 until after the CRC is computed */
	Header.CRC32=CRC.Compute(&Header,sizeof(Header));
	Header.FileSize+=sizeof(Header);

	ofs.seekp(sizeof(Header));	/* Skip the header for now */
	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error processing "+Filename,Exception::LEVEL_ERROR);
	}

	CancelFlag=false;
	SaveAirports(&Airports,&ofs,&(Header.FileSize),&(Header.CRC32));
	SaveNavAids(&NavAids,&ofs,&Header.FileSize,&Header.CRC32);
	SaveEnrouteWaypoints(&EnrouteWaypoints,&ofs,&Header.FileSize,&Header.CRC32);
	SaveGeoCharts(&GeoCharts,&ofs,&Header.FileSize,&Header.CRC32);
	Header.CRC32=CRC.Finalize(Header.CRC32);
	ofs.seekp(0);	/* Write the header now that we have all the details */
	ofs.write((char*)&Header,sizeof(Header));
	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error writing "+Filename,Exception::LEVEL_ERROR);
	}

	ofs.close();

	/* If we just saved, we must be loaded.  The allows the loaded flag to be
	 * set after importing data.
	 */
	Loaded=true;

	/* Update our counts in case they've changed */
	UpdateCounts();
}

void Volume::UpdateCounts() {
	nAirports=Airports.size();
	nNavAids=NavAids.size();
	nEnrouteWaypoints=EnrouteWaypoints.size();
	nGeoCharts=GeoCharts.size();
	nApproachPlates=GetnApproachPlates();
}

void Volume::SaveAirports(std::vector<Airport*> *AirportV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<Airport*>::iterator Airporti;
	std::vector<Runway*> *Runways;
	std::vector<NavAid*> *NavAids;
	std::vector<Frequency*> *Freqs;
	std::vector<ApproachPlate*> *ApproachPlates;
	struct Volume_Record_struct Record;
	ssize_t Size;

	Record.Type=RECORDTYPE_AIRPORT;
	Size=RecordSize(&Record);
	for (Airporti=AirportV->begin();Airporti!=AirportV->end() && !CancelFlag;Airporti++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_AIRPORT;
		if (!((*Airporti)->MakeBinary(&(Record.Airport)))) {
			ofs->close();
			throw Exception("Error preparing airport data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing airport data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;

		Runways=(*Airporti)->GetRunways();
		if (Runways->size())
			SaveRunways(Runways,ofs,RunningSize,RunningCRC);
		NavAids=(*Airporti)->GetNavAids();
		if (NavAids->size())
			SaveNavAids(NavAids,ofs,RunningSize,RunningCRC);
		Freqs=(*Airporti)->GetFreqs();
		if (Freqs->size())
			SaveFreqs(Freqs,ofs,RunningSize,RunningCRC);
		ApproachPlates=(*Airporti)->GetApproachPlates();
		if (ApproachPlates->size())
			SaveApproachPlates(ApproachPlates,ofs,RunningSize,RunningCRC);
	}
}

void Volume::SaveRunways(std::vector<Runway*> *RunwayV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<Runway*>::iterator Runwayi;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_RUNWAY;
	Size=RecordSize(&Record);
	for (Runwayi=RunwayV->begin();Runwayi!=RunwayV->end() && !CancelFlag;Runwayi++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_RUNWAY;
		if (!((*Runwayi)->MakeBinary(&(Record.Runway)))) {
			ofs->close();
			throw Exception("Error preparing runway data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing runway data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
	}
}

void Volume::SaveFreqs(std::vector<Frequency*> *FreqV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<Frequency*>::iterator Freqi;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_FREQ;
	Size=RecordSize(&Record);
	for (Freqi=FreqV->begin();Freqi!=FreqV->end() && !CancelFlag;Freqi++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_FREQ;
		if (!((*Freqi)->MakeBinary(&(Record.Freq)))) {
			ofs->close();
			throw Exception("Error preparing frequency data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing frequency data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
	}
}

void Volume::SaveNavAids(std::vector<NavAid*> *NavAidV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<NavAid*>::iterator NavAidi;
	std::vector<Frequency*> Freqs;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_NAVAID;
	Size=RecordSize(&Record);
	for (NavAidi=NavAidV->begin();NavAidi!=NavAidV->end() && !CancelFlag;NavAidi++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_NAVAID;
		if (!((*NavAidi)->MakeBinary(&(Record.NavAid)))) {
			ofs->close();
			throw Exception("Error preparing navaid data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing navaid data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
		Freqs.push_back((*NavAidi)->GetFreq());
		SaveFreqs(&Freqs,ofs,RunningSize,RunningCRC);
		Freqs.clear();
	}
}

void Volume::SaveEnrouteWaypoints(std::vector<EnrouteWaypoint*> *EnrouteWaypointV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<EnrouteWaypoint*>::iterator EnrouteWaypointi;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_ENROUTEWAYPOINT;
	Size=RecordSize(&Record);
	for (EnrouteWaypointi=EnrouteWaypointV->begin();EnrouteWaypointi!=EnrouteWaypointV->end() && !CancelFlag;EnrouteWaypointi++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_ENROUTEWAYPOINT;
		if (!((*EnrouteWaypointi)->MakeBinary(&(Record.EnrouteWaypoint)))) {
			ofs->close();
			throw Exception("Error preparing enroute waypoing data",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing enroute waypoing data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
	}
}

void Volume::SaveApproachPlates(std::vector<ApproachPlate*> *ApproachPlateV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<ApproachPlate*>::iterator ApproachPlatei;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_APPROACHPLATE;
	Size=RecordSize(&Record);
	for (ApproachPlatei=ApproachPlateV->begin();ApproachPlatei!=ApproachPlateV->end() && !CancelFlag;ApproachPlatei++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_APPROACHPLATE;
		if (!((*ApproachPlatei)->MakeBinary(&(Record.ApproachPlate)))) {
			ofs->close();
			throw Exception("Error preparing approach plate",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing approach plate data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
	}
}

void Volume::SaveGeoCharts(std::vector<GeoChart*> *GeoChartV, std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC) {
	std::vector<GeoChart*>::iterator GeoCharti;
	struct Volume_Record_struct Record;
	int Size;

	Record.Type=RECORDTYPE_GEOCHART;
	Size=RecordSize(&Record);
	for (GeoCharti=GeoChartV->begin();GeoCharti!=GeoChartV->end() && !CancelFlag;GeoCharti++) {
		memset(&Record,0,sizeof(Record));
		Record.Type=RECORDTYPE_GEOCHART;
		if (!((*GeoCharti)->MakeBinary(&(Record.GeoChart)))) {
			ofs->close();
			throw Exception("Error preparing chart",Exception::LEVEL_ERROR);
		}
		*RunningCRC=CRC.Compute(&Record,Size,*RunningCRC);
		ofs->write((char*)&Record,Size);
		if (!ofs->good()) {
			ofs->close();
			throw Exception("Error writing chart data",Exception::LEVEL_ERROR);
		}
		*RunningSize+=Size;
	}
}

void Volume::LoadAirport(std::vector<Airport*> *AirportV,std::ifstream *ifs, uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	Airport *AP;
	int i;

	Record.Type=RECORDTYPE_AIRPORT;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading airport data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_AIRPORT) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	AP=new Airport();
	if (!AP) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	AP->SetFromBinary(&(Record.Airport));

	for (i=0;i<Record.Airport.nRunways && !CancelFlag;i++)
		LoadRunway(AP->GetRunways(),ifs,RunningCRC);
	for (i=0;i<Record.Airport.nNavAids && !CancelFlag;i++)
		LoadNavAid(AP->GetNavAids(),ifs,RunningCRC);
	for (i=0;i<Record.Airport.nFreqs && !CancelFlag;i++)
		LoadFreq(AP->GetFreqs(),ifs,RunningCRC);
	for (i=0;i<Record.Airport.nApproachPlates && !CancelFlag;i++)
		LoadApproachPlate(AP->GetApproachPlates(),AP,ifs,RunningCRC);
	AirportV->push_back(AP);
}

void Volume::LoadRunway(std::vector<Runway*> *RunwayV,std::ifstream *ifs, uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	Runway *R;

	Record.Type=RECORDTYPE_RUNWAY;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading runway data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_RUNWAY) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	R=new Runway();
	if (!R) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	R->SetFromBinary(&(Record.Runway));

	RunwayV->push_back(R);
}

void Volume::LoadFreq(std::vector<Frequency*> *FreqV,std::ifstream *ifs, uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	Frequency *F;

	Record.Type=RECORDTYPE_FREQ;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading frequency data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_FREQ) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	F=new Frequency();
	if (!F) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	F->SetFromBinary(&(Record.Freq));

	FreqV->push_back(F);
}

void Volume::LoadNavAid(std::vector<NavAid*> *NavAidV, std::ifstream *ifs, uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	NavAid *N;

	Record.Type=RECORDTYPE_NAVAID;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading navaid data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_NAVAID) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	switch ((NavAid::Type_enum)Record.NavAid.Type) {
		case NavAid::TYPE_VOR:
			N=new VOR();
			break;
		case NavAid::TYPE_NDB:
			N=new NDB();
			break;
		case NavAid::TYPE_ILS:
			N=new ILS();
			break;
		case NavAid::TYPE_LOCALIZER:
			N=new Localizer();
			break;
	}

	if (!N) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	N->SetFromBinary(&(Record.NavAid));

	std::vector<Frequency*> Freqs;
	std::vector<Frequency*>::iterator Freqi;
	LoadFreq(&Freqs,ifs,RunningCRC);
	Freqi=Freqs.begin();
	N->SetFreq(*Freqi);
	delete *Freqi;

	NavAidV->push_back(N);
}

void Volume::LoadEnrouteWaypoint(std::vector<EnrouteWaypoint*> *EnrouteWaypointV, std::ifstream *ifs, uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	EnrouteWaypoint *E;

	Record.Type=RECORDTYPE_ENROUTEWAYPOINT;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading enroute waypoint data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_ENROUTEWAYPOINT) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	E=new EnrouteWaypoint();
	if (!E) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	E->SetFromBinary(&(Record.EnrouteWaypoint));

	EnrouteWaypointV->push_back(E);
}

void Volume::LoadApproachPlate(
	std::vector<ApproachPlate*> *ApproachPlateV,
	Airport *AP,
	std::ifstream *ifs,
	uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	ApproachPlate *A;

	Record.Type=RECORDTYPE_APPROACHPLATE;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading approach plate data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_APPROACHPLATE) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	A=new ApproachPlate();
	if (!A) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	A->SetFromBinary(&(Record.ApproachPlate));

	if (AP) {
		A->SetAirportIdent(AP->GetID());
		A->SetAirportICAOIdent(AP->GetICAO());
	}

	ApproachPlateV->push_back(A);
}

void Volume::LoadGeoChart(
	std::vector<GeoChart*> *GeoChartV,
	std::ifstream *ifs,
	uint32_t *RunningCRC) {
	struct Volume_Record_struct Record;
	GeoChart *GC;

	Record.Type=RECORDTYPE_GEOCHART;
	ifs->read((char*)&Record,RecordSize(&Record));
	if (!ifs->good()) {
		throw Exception("Error reading chart data",Exception::LEVEL_ERROR);
	}
	if (Record.Type!=RECORDTYPE_GEOCHART) {
		throw Exception("Unexpected record type",Exception::LEVEL_ERROR);
	}
	*RunningCRC=CRC.Compute(&Record,RecordSize(&Record),*RunningCRC);

	GC=new GeoChart();
	if (!GC) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}
	GC->SetFromBinary(&(Record.GeoChart));

	GeoChartV->push_back(GC);
}

int Volume::RecordSize(struct Volume_Record_struct *Record) {
	int Overhead;

	Overhead=sizeof(Record->Type);
	switch ((RecordType_enum)Record->Type) {
		case RECORDTYPE_AIRPORT:
			return Overhead+sizeof(Record->Airport);
		case RECORDTYPE_RUNWAY:
			return Overhead+sizeof(Record->Runway);
		case RECORDTYPE_FREQ:
			return Overhead+sizeof(Record->Freq);
		case RECORDTYPE_NAVAID:
			return Overhead+sizeof(Record->NavAid);
		case RECORDTYPE_ENROUTEWAYPOINT:
			return Overhead+sizeof(Record->EnrouteWaypoint);
		case RECORDTYPE_APPROACHPLATE:
			return Overhead+sizeof(Record->ApproachPlate);
		case RECORDTYPE_GEOCHART:
			return Overhead+sizeof(Record->GeoChart);
	}
	return 0;	/* This should never be reached */
}

void Volume::GetSectionalsByBounds(double Lon, double Lat, std::vector<GeoChart*> *MatchingSectionals)
{
	MatchingSectionals->clear();
	for (std::vector<GeoChart*>::iterator GeoCharti=GeoCharts.begin(); GeoCharti!=GeoCharts.end(); GeoCharti++)
	{
		if (	(*GeoCharti)->GetGeoReferenced() &&
				Lon>(*GeoCharti)->GetWestBounding() &&
				Lon<(*GeoCharti)->GetEastBounding() &&
				Lat>(*GeoCharti)->GetSouthBounding() &&
				Lat<(*GeoCharti)->GetNorthBounding()
				)
		{
			MatchingSectionals->push_back(*GeoCharti);
		}
	}
}

GeoChart* Volume::GetBestSectional(double Lon, double Lat, std::vector<GeoChart*> *Choices)
{
	std::vector<GeoChart*> LocalChoices,TACChoices;
	std::vector<GeoChart*>::iterator iter;
	double MinEastBounding=400.0;	/* Impossibly high since it's greater than 360 degrees */

	if (!Choices)
	{
		GetSectionalsByBounds(Lon,Lat,&LocalChoices);
		Choices=&LocalChoices;
	}

	if (Choices->size()==0)
		return NULL;

	/* If any TAC's are available, don't consider the Sectionals */
	for (iter=Choices->begin();iter!=Choices->end();++iter) {
		if ((*iter)->GetType()==GeoChart::TYPE_TAC) {
			TACChoices.push_back(*iter);
		}
	}
	if (TACChoices.size()) {
		Choices=&TACChoices;
	}

	/* Find the greatest East Bounding since the key is on the left of the chart */
	for (iter=Choices->begin();iter!=Choices->end();++iter) {
		if ((*iter)->GetEastBounding()<MinEastBounding) {
			MinEastBounding=(*iter)->GetEastBounding();
		}
	}

	/* Use the chart that's furthest East */
	for (iter=Choices->begin();iter!=Choices->end();++iter) {
		if ((*iter)->GetEastBounding()==MinEastBounding) {
			return (*iter);
		}
	}

	/* Just return the first one as a last resort */
	return *(Choices->begin());
}

void Volume::on_CancelButton_clicked() {
	CancelFlag=true;
}

std::string Volume::GetDataDir() {
	return Cfg->GetDataStoreDir()+"/"+Name;
}

bool Volume::GetLoaded() {
	return Loaded;
}

/* Load just the header data */
bool Volume::Scan(std::string Filename, struct Volume_Header_struct *Header) {
	std::ifstream ifs;
	struct stat statdata;
	struct Volume_Header_struct TmpHeader;

	Util::ShellExpand(&Filename);

	if (Filename.compare("")!=0) {
		this->Filename=Filename;
	} else {
		Filename=this->Filename;
	}

	if (stat(Filename.c_str(),&statdata)!=0) {
		if (errno==ENOENT) {
			return false;
		}

		throw Exception("Error processing "+Filename,Exception::LEVEL_ERROR);
	}

	if (!Header)
		Header=&TmpHeader;

	ifs.open(Filename.c_str(),std::ifstream::binary);
	if (!ifs.good()) {
		throw Exception("Error opening "+Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	/* Confirm the header is valid other than the CRC */
	ifs.read((char*)Header,sizeof(*Header));
	if (!ifs.good()) {
		throw Exception("File corrupt: cannot read header of "+Filename,Exception::LEVEL_ERROR);
	}
	if (strncmp(Header->Magic,VOLUME_FILEMAGIC,sizeof(VOLUME_FILEMAGIC)!=0)) {
		throw Exception("File corrupt: bad magic in "+Filename,Exception::LEVEL_ERROR);
	}
	if (Header->DataVersion!=VOLUME_HEADER_STRUCT_VER) {
		throw Exception("Unsupported data version in "+Filename,Exception::LEVEL_ERROR);
	}
	/* Skip effective from/to for now */
	if (Header->FileSize!=(unsigned int)statdata.st_size) {
		throw Exception("File corrupt: size mismatch in "+Filename,Exception::LEVEL_ERROR);
	}

	/* Store the vital stats */
	Name.assign(Header->VolumeName,sizeof(Header->VolumeName));
	Name=Name.c_str();	/* Reset the size to the actual number of characters */
	EffectiveFrom=Header->EffectiveFrom;
	EffectiveTo=Header->EffectiveTo;
	nAirports=Header->nAirports;
	nNavAids=Header->nNavAids;
	nEnrouteWaypoints=Header->nEnrouteWaypoints;
	nGeoCharts=Header->nGeoCharts;
	nApproachPlates=Header->nApproachPlates;

	return true;
}

time_t Volume::ComputeEffectiveFrom(std::string VolStr) {
	int EffYear, Edition, Editioni=1, PrevYear=2000;
	time_t Timestamp;
	struct tm bdtime;

	if (VolStr.size()!=4)
		return 0;
	EffYear=atoi(VolStr.substr(0,2).c_str());
	if (EffYear<=0 || EffYear>99)
		return 0;
	Edition=atoi(VolStr.substr(2,2).c_str());
	if (Edition<=0 || Edition>13)
		return 0;

	/* Initialize bdtime */
	Timestamp=time(NULL);

	try {
		Util::GmTime_r(&bdtime,Timestamp);
	}
	catch (Exception e) {
		e.Show();
		return false;
	}

	/* Set the start date/time - assume Volume 0001 became effective Jan 27, 2000 */
	bdtime.tm_year=100;	/* Year 2000 */
	bdtime.tm_mon=0;
	bdtime.tm_mday=27;
	bdtime.tm_hour=9;
	bdtime.tm_min=1;
	bdtime.tm_sec=0;

	while (!(bdtime.tm_year-100==EffYear && Editioni==Edition) && bdtime.tm_year!=199) {
		Timestamp=Util::GMMkTime(&bdtime)+28*24*60*60;	/* Go forward 28 days */

		try {
			Util::GmTime_r(&bdtime,Timestamp);
		}
		catch (Exception e) {
			e.Show();
			return false;
		}

		Editioni++;
		if (bdtime.tm_year!=PrevYear-1900) {
			PrevYear=bdtime.tm_year+1900;
			Editioni=1;
		}
	}

	if (bdtime.tm_year==199)	/* No suitable cycle was found */
		return 0;

	return Timestamp;
}

time_t Volume::ComputeEffectiveTo(std::string VolStr) {
	/* Valid for 28 days.  Subtract 1 second so there's no overlap */
	return ComputeEffectiveFrom(VolStr)+28*24*60*60-1;
}

std::string Volume::GetDirectory() {
	std::string Dir;
	Dir=Cfg->GetDataStoreDir()+"/"+Name;
	Util::ShellExpand(&Dir);
	return Dir;
}

bool Volume::IsPlateImageFileUsed(std::string Filename) {
	for (std::vector<Airport*>::iterator Airporti=Airports.begin();Airporti!=Airports.end();++Airporti) {
		for (std::vector<ApproachPlate*>::iterator ApproachPlatei=(*Airporti)->GetApproachPlates()->begin();ApproachPlatei!=(*Airporti)->GetApproachPlates()->end() && !CancelFlag;++ApproachPlatei) {
			if ((*ApproachPlatei)->GetImgName()==Filename)
				return true;
		}
	}
	return false;
}

DataStore::DataStore(std::string *GladeFilename, Config *Cfg)
{
	this->GladeFilename=GladeFilename;
	this->Cfg=Cfg;
	ActiveVolume=Volumes.end();
}

DataStore::~DataStore()
{
	Clear();
}

void DataStore::Clear()
{
	for(std::list<Volume*>::iterator Volumei = Volumes.begin(); Volumei != Volumes.end(); ++Volumei)
	{
		delete *Volumei;
	}
	Volumes.clear();
}

void DataStore::ClearVolumes()
{
	for(std::list<Volume*>::iterator Volumei = Volumes.begin(); Volumei != Volumes.end(); ++Volumei)
	{
		(*Volumei)->Clear();
	}
}

void DataStore::ClearUnloadedVolumes()
{
	for(std::list<Volume*>::iterator Volumei = Volumes.begin(); Volumei != Volumes.end(); ++Volumei)
	{
		if (Volumei!=ActiveVolume) {
			(*Volumei)->Clear();
		}
	}
}

Airport *DataStore::GetAirportByICAO(std::string ICAO) {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetAirportByICAO(ICAO);
}

Airport *DataStore::GetAirportByIATA(std::string IATA) {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetAirportByIATA(IATA);
}

std::vector<Airport*> *DataStore::GetAirports() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetAirports();
}

unsigned int DataStore::GetnAirports() {
	if (ActiveVolume==Volumes.end()) {
		return 0;
	}

	return (*ActiveVolume)->GetnAirports();
}

std::vector<NavAid*> *DataStore::GetNavAids() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetNavAids();
}

unsigned int DataStore::GetnNavAids() {
	if (ActiveVolume==Volumes.end()) {
		return 0;
	}

	return (*ActiveVolume)->GetnNavAids();
}

std::vector<EnrouteWaypoint*> *DataStore::GetEnrouteWaypoints() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetEnrouteWaypoints();
}

unsigned int DataStore::GetnEnrouteWaypoints() {
	if (ActiveVolume==Volumes.end()) {
		return 0;
	}

	return (*ActiveVolume)->GetnEnrouteWaypoints();
}

std::vector<GeoChart*> *DataStore::GetGeoCharts() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetGeoCharts();
}

unsigned int DataStore::GetnGeoCharts() {
	if (ActiveVolume==Volumes.end()) {
		return 0;
	}

	return (*ActiveVolume)->GetnGeoCharts();
}

unsigned int DataStore::GetnApproachPlates() {
	if (ActiveVolume==Volumes.end()) {
		return 0;
	}

	return (*ActiveVolume)->GetnApproachPlates();
}

void DataStore::WaypointSearch(std::string Str, std::vector<Waypoint*> *Waypoints) {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	(*ActiveVolume)->WaypointSearch(Str,Waypoints);
}

time_t DataStore::GetEffectiveFrom() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetEffectiveFrom();
}

time_t DataStore::GetEffectiveTo() {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetEffectiveTo();
}

void DataStore::GetSectionalsByBounds(double Lon, double Lat, std::vector<GeoChart*> *MatchingSectionals) {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	(*ActiveVolume)->GetSectionalsByBounds(Lon,Lat,MatchingSectionals);
}

GeoChart* DataStore::GetBestSectional(double Lon, double Lat, std::vector<GeoChart*> *Choices) {
	if (ActiveVolume==Volumes.end()) {
		throw Exception("No data selected in datastore",Exception::LEVEL_ERROR);
	}

	return (*ActiveVolume)->GetBestSectional(Lon,Lat,Choices);
}

bool DataStore::HasActiveVolume() {
	return ActiveVolume!=Volumes.end();
}

std::list<Volume*>::iterator DataStore::GetActiveVolume() {
	return ActiveVolume;
}

bool DataStore::SetActiveVolume(std::list<Volume*>::iterator Volumei) {
	std::ostringstream OStr,OTimeStr;
	time_t timediff;

	if (Volumei==ActiveVolume)
		return true;	/* Nothing to do */

	timediff=(*Volumei)->GetEffectiveFrom()-time(NULL);
	if (timediff>0) {
		if (timediff<60) {
			OTimeStr<<timediff<<" second(s)";
		} else if (timediff<(60*60)) {
			OTimeStr<<(timediff/60)<<" minute(s)";
		} else if (timediff<(24*60*60)) {
			OTimeStr<<(timediff/(60*60))<<" hour(s)";
		} else {
			OTimeStr<<(timediff/(24*60*60))<<" day(s)";
		}
		OStr<<"This Data Volume will not be effective for "<<
			OTimeStr.str()<<".  Continue?";
		Gtk::MessageDialog *Message = new Gtk::MessageDialog(OStr.str(),
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK_CANCEL,true);
		if (Message->run()!=Gtk::RESPONSE_OK) {
			delete Message;
			return false;
		}
		delete Message;
	}

	OStr.clear();
	OTimeStr.clear();
	timediff=time(NULL)-(*Volumei)->GetEffectiveTo();
	if (timediff>0) {
		if (timediff<60) {
			OTimeStr<<timediff<<" second(s)";
		} else if (timediff<(60*60)) {
			OTimeStr<<(timediff/60)<<" minute(s)";
		} else if (timediff<(24*60*60)) {
			OTimeStr<<(timediff/(60*60))<<" hour(s)";
		} else {
			OTimeStr<<(timediff/(24*60*60))<<" day(s)";
		}
		OStr<<"This Data Volume has not be effective for "<<
			OTimeStr.str()<<".  Continue?";
		Gtk::MessageDialog *Message = new Gtk::MessageDialog(OStr.str(),
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK_CANCEL,true);
		if (Message->run()!=Gtk::RESPONSE_OK) {
			delete Message;
			return false;
		}
		delete Message;
	}

	ClearVolumes();	/* Free old data since we're about to load new */
	if (!(*Volumei)->GetLoaded()) {
		(*Volumei)->Load();
	}
	ActiveVolume=Volumei;

	return true;
}

Volume *DataStore::GetVolume(std::string *Name, time_t *EffectiveFrom, time_t *EffectiveTo) {
	for (std::list<Volume*>::iterator Volumei=Volumes.begin();Volumei!=Volumes.end();++Volumei) {
		if (
			(!Name || (*Volumei)->GetName().compare(*Name)==0) &&
			(!EffectiveFrom || (*Volumei)->GetEffectiveFrom()<=*EffectiveFrom) &&
			(!EffectiveTo || (*Volumei)->GetEffectiveTo()>=*EffectiveTo)
			) {
			if (!(*Volumei)->GetLoaded()) {
				(*Volumei)->Load();
			}
			return (*Volumei);
		}
	}

	if (Name && EffectiveFrom && EffectiveTo) {
		/* Create a new volume */
		Volume *NewVol=new Volume(GladeFilename,Cfg,*Name,*EffectiveFrom,*EffectiveTo);

		if (!NewVol) {
			throw Exception("Unable to allocate memory",Exception::LEVEL_ERROR);
		}
		Volumes.push_back(NewVol);

		return NewVol;
	}

	return NULL;
}

void DataStore::ScanVolumes() {
	DIR *dir;
	dirent *ent;
	std::string Path;

	Clear();

	Path=Cfg->GetDataStoreDir();
	Util::ShellExpand(&Path);

	errno=0;
	dir=opendir(Path.c_str());
	if (!dir) {
		throw Exception("Error opening directory: "+Path+" ("+
			strerror(errno)+")",Exception::LEVEL_ERROR);
	}

	do {
		/* Copy all other files that are associated with any loaded approach at any loaded airport */
		ent=readdir(dir);
#ifndef WIN32
/* We can do the right thing and only descend into directories */
		if (ent && ent->d_name[0]!='.' && ent->d_type==DT_DIR) {
#else
/* We have to assume that any entry without a period in it is a directory */
		if (ent && strchr(ent->d_name,'.')==NULL) {
#endif
			Volume *NewVol=new Volume(GladeFilename,Cfg);
			if (!NewVol) {
				closedir(dir);
				throw Exception(_("Cannot allocate memory"),Exception::LEVEL_ERROR);
			}

			if (!NewVol->Scan(Path+"/"+ent->d_name+"/volume.vifir")) {
				errno=0;
				delete NewVol;
				NewVol=NULL;
			} else {
				Volumes.push_back(NewVol);
			}

		}
	} while (ent);
	closedir(dir);
	if (errno) {
		throw Exception("Error reading volume dir: "+Path+" ("+
			strerror(errno)+")",Exception::LEVEL_ERROR);
	}
}

void DataStore::ActivateCurrentVolume() {
	ActiveVolume=Volumes.end();	/* Select nothing */
	for (std::list<Volume*>::iterator Volumei=Volumes.begin();Volumei!=Volumes.end();++Volumei) {
		if ((*Volumei)->GetEffectiveFrom()<=time(NULL) && (*Volumei)->GetEffectiveTo()>=time(NULL)) {
			SetActiveVolume(Volumei);
		}
	}
}

std::list<Volume*> *DataStore::GetVolumes() {
	return &Volumes;
}
