/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <stdlib.h>
#include <math.h>
#include <cstring>
#include "NavAid.h"
#include "ARINC424_18.h"

NavAid::NavAid()
{
	LocType=Waypoint::LOC_NAVAID;
	Declination=0.0;
}

NavAid::~NavAid()
{
}

Frequency *NavAid::GetFreq()
{
	return &Freq;
}

void NavAid::SetFreq(Frequency *NewFreq)
{
	Freq.Clone(NewFreq);
}

bool NavAid::ParseNFD(std::string Line)
{
	struct arinc424_18_navaid_header_struct *NFDHeader;
	char strbuf[sizeof(struct arinc424_18_navaid_header_struct)];

	if (Line.length()!=sizeof(struct arinc424_18_vhfnavaid_struct)+1)	/* Allow for LF - all the lines are the same length */
	{
		return false;
	}

	if (
		strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_NAVAID_VHF)!=0 &&
		strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_NAVAID_NDB)!=0
		)
	{
		return false;
	}
	NFDHeader=(struct arinc424_18_navaid_header_struct *)Line.c_str();

	/* ID */
	memcpy(strbuf,NFDHeader->id,sizeof(NFDHeader->id));
	strbuf[sizeof(NFDHeader->id)]=0;

	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	ID=strbuf;

	return true;
}

VOR::VOR()
{
}

enum NavAid::Type_enum VOR::GetType() {
	return TYPE_VOR;
}

bool VOR::ParseNFD(std::string Line)
{
	arinc424_18_vhfnavaid_struct *NFDData;
	char strbuf[sizeof(struct arinc424_18_airport_struct)];

	if (!NavAid::ParseNFD(Line))
	{
		return false;
	}

	NFDData=(struct arinc424_18_vhfnavaid_struct *)Line.c_str();

	/* Lat/Lon */
	Lat=arinc424_18_parse_lat(NFDData->lat);
	Lon=arinc424_18_parse_lon(NFDData->lon);
	if (isnan(Lat) || isnan(Lon))
	{
		return false;
	}

	/* Frequency */
	Freq.SetRawFreq(arinc424_18_parse_freq(NFDData->frequency));
	Freq.SetPrecision(3);
	Freq.SetUnits(Frequency::MHZ);
	Freq.SetType(Frequency::VOR);

	/* Declination */
	memcpy(strbuf,NFDData->declination,sizeof(NFDData->declination));
	strbuf[sizeof(NFDData->declination)]=0;
	Declination=(atoi(&(strbuf[1])))/10.0;
	if (strbuf[0]=='W')
		Declination=-Declination;

	/* Name */
	memcpy(strbuf,NFDData->name,sizeof(NFDData->name));
	strbuf[sizeof(NFDData->name)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	Name=strbuf;

	return true;
}

NDB::NDB()
{
}

enum NavAid::Type_enum NDB::GetType() {
	return TYPE_NDB;
}

bool NDB::ParseNFD(std::string Line)
{
	arinc424_18_ndbnavaid_struct *NFDData;
	char strbuf[sizeof(struct arinc424_18_airport_struct)];
	if (!NavAid::ParseNFD(Line)) {
		return false;
	}

	NFDData=(struct arinc424_18_ndbnavaid_struct *)Line.c_str();

	Lat=arinc424_18_parse_lat(NFDData->lat);
	Lon=arinc424_18_parse_lon(NFDData->lon);
	if (isnan(Lat) || isnan(Lon))
	{
		return false;
	}

	/* Frequency */
	Freq.SetRawFreq(arinc424_18_parse_freq(NFDData->frequency));
	Freq.SetPrecision(3);
	Freq.SetUnits(Frequency::MHZ);
	Freq.SetType(Frequency::NDB);

	/* Name */
	memcpy(strbuf,NFDData->name,sizeof(NFDData->name));
	strbuf[sizeof(NFDData->name)]=0;
	/* Trim trailing spaces */
	while (isspace(strbuf[strlen(strbuf)-1]) && strlen(strbuf))
		strbuf[strlen(strbuf)-1]=0;
	Name=strbuf;

	return true;
}

bool NDB::ParseXPlane810(std::string *Line) {
	char *tok;
	char *buf;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		/* Ignore the row code */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lat=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lon=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Unused */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Freq.SetType(Frequency::NDB);
		Freq.SetUnits(Frequency::KHZ);
		Freq.SetRawFreq(atoi(tok));
		Freq.SetPrecision(0);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore Minimum range */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Unused */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		ID=tok;
		if (!(tok=strtok(NULL,"")))
			throw std::exception();
		Name=tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	free(buf);
	return true;
}

bool VOR::ParseXPlane810(std::string *Line) {
	char *tok;
	char *buf;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		/* Ignore the row code */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lat=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lon=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore elevation */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Freq.SetType(Frequency::VOR);
		Freq.SetUnits(Frequency::MHZ);
		Freq.SetRawFreq(roundf(1000*atoi(tok)));
		Freq.SetPrecision(3);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore Minimum range */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore slaved variation */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		ID=tok;
		if (!(tok=strtok(NULL,"")))
			throw std::exception();
		Name=tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	free(buf);
	return true;
}

bool ILS::ParseXPlane810(std::string *Line) {
	char *tok;
	char *buf;
	std::string Airport,Runway;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		/* Ignore the row code */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lat=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lon=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore elevation */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Freq.SetType(Frequency::VOR);
		Freq.SetUnits(Frequency::MHZ);
		Freq.SetRawFreq(roundf(1000*atoi(tok)));
		Freq.SetPrecision(3);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore Minimum range */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore true bearing */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		ID=tok;
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Airport=tok;
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Runway=tok;
		if (!(tok=strtok(NULL,"")))
			throw std::exception();
		Name=Airport+" runway "+Runway+" "+tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	free(buf);
	return true;
}

bool Localizer::ParseXPlane810(std::string *Line) {
	char *tok;
	char *buf;
	std::string Airport,Runway;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	try {
		/* Parse the fields */
		if (!(tok=strtok(buf," \t")))
			throw std::exception();
		/* Ignore the row code */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lat=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Lon=atof(tok);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore elevation */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Freq.SetType(Frequency::VOR);
		Freq.SetUnits(Frequency::MHZ);
		Freq.SetRawFreq(roundf(1000*atoi(tok)));
		Freq.SetPrecision(3);
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore Minimum range */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		/* Ignore true bearing */
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		ID=tok;
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Airport=tok;
		if (!(tok=strtok(NULL," \t")))
			throw std::exception();
		Runway=tok;
		if (!(tok=strtok(NULL,"")))
			throw std::exception();
		Name=Airport+" runway "+Runway+" "+tok;
	}

	catch (...) {
		free(buf);
		return false;
	}

	free(buf);
	return true;
}

ILS::ILS()
{
}

enum NavAid::Type_enum ILS::GetType() {
	return TYPE_ILS;
}

Localizer::Localizer()
{
}

enum NavAid::Type_enum Localizer::GetType() {
	return TYPE_LOCALIZER;
}

bool NavAid::MakeBinary(struct NavAid_Binary_struct *BinaryData) {
	double Lonx1000000,Latx1000000,Decx100;

	Lonx1000000=Lon*1000000.0;
	Latx1000000=Lat*1000000.0;
	Decx100=Declination*100.0;
	/* Sanity checking */
	if (
		ID.size()>sizeof(BinaryData->ID) ||
		Lonx1000000<INT32_MIN || Lonx1000000>INT32_MAX ||
		Latx1000000<INT32_MIN || Latx1000000>INT32_MAX ||
		Decx100>INT32_MAX || Decx100<INT32_MIN ||
		LocType>INT8_MAX ||
		Name.size()>sizeof(BinaryData->Name)
		) {
		return false;
	}

	strncpy(BinaryData->ID,ID.c_str(),sizeof(BinaryData->ID));
	BinaryData->Lonx1000000=(int32_t)round(Lonx1000000);
	BinaryData->Latx1000000=(int32_t)round(Latx1000000);
	BinaryData->Declinationx100=(int32_t)round(Decx100);
	BinaryData->Type=GetType();
	strncpy(BinaryData->Name,Name.c_str(),sizeof(BinaryData->Name));
	return true;
}

void NavAid::SetFromBinary(struct NavAid_Binary_struct *BinaryData) {
	ID.assign(BinaryData->ID,sizeof(BinaryData->ID));
	Lon=BinaryData->Lonx1000000/1000000.0;
	Lat=BinaryData->Latx1000000/1000000.0;
	Declination=BinaryData->Declinationx100/100.0;
	Name.assign(BinaryData->Name,sizeof(BinaryData->Name));
	/* Reset the size to the actual number of characters */
	ID=ID.c_str();
	Name=Name.c_str();

	/* The Type is set when the class is instantiated */
}
